@@ -9,23 +9,17 @@ import functools
 import itertools
 import hashlib
 import weakref
+import queue
 from collections import defaultdict
-try:
-    # Python 3.8+ https://docs.python.org/3/whatsnew/3.7.html#id3
-    from collections.abc import MutableMapping
-except ImportError:
-    # Python 2.6, 2.7
-    from collections import MutableMapping
+from collections.abc import MutableMapping
 
 import six
-from six.moves import queue
 import redis
 
 from ._zset import ZSet
 
 
 MAX_STRING_SIZE = 512 * 1024 * 1024
-INF = float('inf')
 
 INVALID_EXPIRE_MSG = "invalid expire time in {}"
 WRONGTYPE_MSG = \
@@ -47,6 +41,7 @@ INVALID_SORT_FLOAT_MSG = "One or more scores can't be converted into double"
 SRC_DST_SAME_MSG = "source and destination objects are the same"
 NO_KEY_MSG = "no such key"
 INDEX_ERROR_MSG = "index out of range"
+ZADD_NX_XX_ERROR_MSG = "ZADD allows either 'nx' or 'xx', not both"
 ZUNIONSTORE_KEYS_MSG = "at least 1 input key is needed for ZUNIONSTORE/ZINTERSTORE"
 WRONG_ARGS_MSG = "wrong number of arguments for '{}' command"
 UNKNOWN_COMMAND_MSG = "unknown command '{}'"
@@ -68,13 +63,13 @@ CONNECTION_ERROR_MSG = "FakeRedis is emulating a connection error."
 FLAG_NO_SCRIPT = 's'      # Command not allowed in scripts
 
 
-class SimpleString(object):
+class SimpleString:
     def __init__(self, value):
         assert isinstance(value, bytes)
         self.value = value
 
 
-class NoResponse(object):
+class NoResponse:
     """Returned by pub/sub commands to indicate that no response should be returned"""
     pass
 
@@ -85,25 +80,6 @@ PONG = SimpleString(b'PONG')
 BGSAVE_STARTED = SimpleString(b'Background saving started')
 
 
-if six.PY2:
-    def isfinite(value):
-        return not math.isinf(value) and not math.isnan(value)
-
-    # Not the same as six.byte2int, which takes a bytes in both Python 2+3.
-    # This takes an integer in Python 3.
-    def byte_to_int(value):
-        return ord(value)
-else:
-    def isfinite(value):
-        return math.isfinite(value)
-
-    def byte_to_int(value):
-        assert isinstance(value, int)
-        return value
-
-    long = int
-
-
 def null_terminate(s):
     # Redis uses C functions on some strings, which means they stop at the
     # first NULL.
@@ -187,7 +163,7 @@ def compile_pattern(pattern):
     return re.compile(regex, re.S)
 
 
-class Item(object):
+class Item:
     """An item stored in the database"""
 
     __slots__ = ['value', 'expireat']
@@ -197,7 +173,7 @@ class Item(object):
         self.expireat = None
 
 
-class CommandItem(object):
+class CommandItem:
     """An item referenced by a command.
 
     It wraps an Item but has extra fields to manage updates and notifications.
@@ -333,7 +309,7 @@ class Hash(dict):
     redis_type = b'hash'
 
 
-class Int(object):
+class Int:
     """Argument converter for 64-bit signed integers"""
 
     DECODE_ERROR = INVALID_INT_MSG
@@ -349,7 +325,7 @@ class Int(object):
     def decode(cls, value):
         try:
             out = int(value)
-            if not cls.valid(out) or six.ensure_binary(str(out)) != value:
+            if not cls.valid(out) or str(out).encode() != value:
                 raise ValueError
         except ValueError:
             raise redis.ResponseError(cls.DECODE_ERROR)
@@ -358,7 +334,7 @@ class Int(object):
     @classmethod
     def encode(cls, value):
         if cls.valid(value):
-            return six.ensure_binary(str(value))
+            return str(value).encode()
         else:
             raise redis.ResponseError(cls.ENCODE_ERROR)
 
@@ -392,7 +368,7 @@ class Timeout(Int):
     MIN_VALUE = 0
 
 
-class Float(object):
+class Float:
     """Argument converter for floating-point values.
 
     Redis uses long double for some cases (INCRBYFLOAT, HINCRBYFLOAT)
@@ -426,7 +402,7 @@ class Float(object):
                 # Values that over- or underflow- are explicitly rejected by
                 # redis. This is a crude hack to determine whether the input
                 # may have been such a value.
-                if out in (INF, -INF, 0.0) and re.match(b'^[^a-zA-Z]*[1-9]', value):
+                if out in (math.inf, -math.inf, 0.0) and re.match(b'^[^a-zA-Z]*[1-9]', value):
                     raise ValueError
             return out
         except ValueError:
@@ -435,14 +411,14 @@ class Float(object):
     @classmethod
     def encode(cls, value, humanfriendly):
         if math.isinf(value):
-            return six.ensure_binary(str(value))
+            return str(value).encode()
         elif humanfriendly:
             # Algorithm from ld2string in redis
             out = '{:.17f}'.format(value)
             out = re.sub(r'(?:\.)?0+$', '', out)
-            return six.ensure_binary(out)
+            return out.encode()
         else:
-            return six.ensure_binary('{:.17g}'.format(value))
+            return '{:.17g}'.format(value).encode()
 
 
 class SortFloat(Float):
@@ -450,11 +426,11 @@ class SortFloat(Float):
 
     @classmethod
     def decode(cls, value):
-        return super(SortFloat, cls).decode(
+        return super().decode(
             value, allow_leading_whitespace=True, allow_empty=True, crop_null=True)
 
 
-class ScoreTest(object):
+class ScoreTest:
     """Argument converter for sorted set score endpoints."""
     def __init__(self, value, exclusive=False):
         self.value = value
@@ -489,7 +465,7 @@ class ScoreTest(object):
         return (self.value, BeforeAny() if self.exclusive else AfterAny())
 
 
-class StringTest(object):
+class StringTest:
     """Argument converter for sorted set LEX endpoints."""
     def __init__(self, value, exclusive):
         self.value = value
@@ -510,7 +486,7 @@ class StringTest(object):
 
 
 @functools.total_ordering
-class BeforeAny(object):
+class BeforeAny:
     def __gt__(self, other):
         return False
 
@@ -519,7 +495,7 @@ class BeforeAny(object):
 
 
 @functools.total_ordering
-class AfterAny(object):
+class AfterAny:
     def __lt__(self, other):
         return False
 
@@ -527,7 +503,7 @@ class AfterAny(object):
         return isinstance(other, AfterAny)
 
 
-class Key(object):
+class Key:
     """Marker to indicate that argument in signature is a key"""
     UNSPECIFIED = object()
 
@@ -536,7 +512,7 @@ class Key(object):
         self.missing_return = missing_return
 
 
-class Signature(object):
+class Signature:
     def __init__(self, name, fixed, repeat=(), flags=""):
         self.name = name
         self.fixed = fixed
@@ -595,7 +571,7 @@ def valid_response_type(value, nested=False):
     if isinstance(value, NoResponse) and not nested:
         return True
     if value is not None and not isinstance(value, (bytes, SimpleString, redis.ResponseError,
-                                                    int, long, list)):
+                                                    int, list)):
         return False
     if isinstance(value, list):
         if any(not valid_response_type(item, True) for item in value):
@@ -612,7 +588,7 @@ def command(*args, **kwargs):
     return decorator
 
 
-class FakeServer(object):
+class FakeServer:
     def __init__(self):
         self.lock = threading.Lock()
         self.dbs = defaultdict(lambda: Database(self.lock))
@@ -625,7 +601,7 @@ class FakeServer(object):
         self.connected = True
 
 
-class FakeSocket(object):
+class FakeSocket:
     def __init__(self, server):
         self._server = server
         self._db = server.dbs[0]
@@ -725,8 +701,6 @@ class FakeSocket(object):
             return [self._decode_result(r) for r in result]
         elif isinstance(result, SimpleString):
             return result.value
-        elif six.PY2 and isinstance(result, int):
-            return long(result)
         else:
             return result
 
@@ -869,7 +843,8 @@ class FakeSocket(object):
         if pattern is not None:
             regex = compile_pattern(pattern)
             for val in itertools.islice(data, cursor, result_cursor):
-                if regex.match(val):
+                compare_val = val[0] if isinstance(val, tuple) else val
+                if regex.match(compare_val):
                     result_data.append(val)
         else:
             result_data = data[cursor:result_cursor]
@@ -911,8 +886,7 @@ class FakeSocket(object):
     # Key commands
     # TODO: lots
 
-    @command((Key(),), (Key(),), name='del')
-    def del_(self, *keys):
+    def _delete(self, *keys):
         ans = 0
         done = set()
         for key in keys:
@@ -922,6 +896,14 @@ class FakeSocket(object):
                 ans += 1
         return ans
 
+    @command((Key(),), (Key(),), name='del')
+    def del_(self, *keys):
+        return self._delete(*keys)
+
+    @command((Key(),), (Key(),), name='unlink')
+    def unlink(self, *keys):
+        return self._delete(*keys)
+
     @command((Key(),), (Key(),))
     def exists(self, *keys):
         ret = 0
@@ -1007,9 +989,8 @@ class FakeSocket(object):
             raise redis.ResponseError(SRC_DST_SAME_MSG)
         if not key or key.key in self._server.dbs[db]:
             return 0
-        # TODO: what is the interaction with expiry and WATCH?
-        self._server.dbs[db][key.key] = self._server.dbs[self._db_num][key.key]
-        key.value = None   # Causes deletion
+        # TODO: what is the interaction with expiry?
+        self._server.dbs[db][key.key] = self._server.dbs[self._db_num].pop(key.key)
         return 1
 
     @command(())
@@ -1259,7 +1240,7 @@ class FakeSocket(object):
             value = key.value[start:end]
         else:
             value = key.value
-        return sum([bin(byte_to_int(l)).count('1') for l in value])
+        return bin(int.from_bytes(value, 'little')).count('1')
 
     @command((Key(bytes), Int))
     def decrby(self, key, amount):
@@ -1283,7 +1264,7 @@ class FakeSocket(object):
     def incrbyfloat(self, key, amount):
         # TODO: introduce convert_order so that we can specify amount is Float
         c = Float.decode(key.get(b'0')) + Float.decode(amount)
-        if not isfinite(c):
+        if not math.isfinite(c):
             raise redis.ResponseError(NONFINITE_MSG)
         encoded = Float.encode(c, True)
         key.update(encoded)
@@ -1300,7 +1281,7 @@ class FakeSocket(object):
         remaining = offset % 8
         actual_bitoffset = 7 - remaining
         try:
-            actual_val = byte_to_int(value[byte])
+            actual_val = value[byte]
         except IndexError:
             return 0
         return 1 if (1 << actual_bitoffset) & actual_val else 0
@@ -1316,7 +1297,7 @@ class FakeSocket(object):
             # bit.
             needed = byte - (len(val) - 1)
             val += b'\x00' * needed
-        old_byte = byte_to_int(val[byte])
+        old_byte = val[byte]
         if value == 1:
             new_byte = old_byte | (1 << actual_bitoffset)
         else:
@@ -1481,7 +1462,7 @@ class FakeSocket(object):
     @command((Key(Hash), bytes, bytes))
     def hincrbyfloat(self, key, field, amount):
         c = Float.decode(key.value.get(field, b'0')) + Float.decode(amount)
-        if not isfinite(c):
+        if not math.isfinite(c):
             raise redis.ResponseError(NONFINITE_MSG)
         encoded = Float.encode(c, True)
         key.value[field] = encoded
@@ -1737,15 +1718,27 @@ class FakeSocket(object):
     def scard(self, key):
         return len(key.value)
 
-    def _setop(self, op, stop_when_empty, dst, key, *keys):
+    def _calc_setop(self, op, stop_if_missing, key, *keys):
+        if stop_if_missing and not key.value:
+            return set()
         ans = key.value.copy()
         for other in keys:
-            if stop_when_empty and not ans:
-                break
             value = other.value if other.value is not None else set()
             if not isinstance(value, set):
                 raise redis.ResponseError(WRONGTYPE_MSG)
+            if stop_if_missing and not value:
+                return set()
             ans = op(ans, value)
+        return ans
+
+    def _setop(self, op, stop_if_missing, dst, key, *keys):
+        """Apply one of SINTER[STORE], SUNION[STORE], SDIFF[STORE].
+
+        If `stop_if_missing`, the output will be made an empty set as soon as
+        an empty input set is encountered (use for SINTER[STORE]). May assume
+        that `key` is a set (or empty), but `keys` could be anything.
+        """
+        ans = self._calc_setop(op, stop_if_missing, key, *keys)
         if dst is None:
             return list(ans)
         else:
@@ -1760,7 +1753,9 @@ class FakeSocket(object):
     def sdiffstore(self, dst, *keys):
         return self._setop(lambda a, b: a - b, False, dst, *keys)
 
-    @command((Key(set, missing_return=[]),), (Key(),))
+    # The following keys can't be marked as sets because of the
+    # stop_if_empty early-out.
+    @command((Key(set),), (Key(),))
     def sinter(self, *keys):
         return self._setop(lambda a, b: a & b, True, None, *keys)
 
@@ -1895,19 +1890,55 @@ class FakeSocket(object):
 
     @command((Key(ZSet), bytes, bytes), (bytes,))
     def zadd(self, key, *args):
-        # TODO: handle NX, XX, CH, INCR
+        # TODO: handle INCR
         zset = key.value
-        if len(args) % 2 != 0:
+
+        i = 0
+        ch = False
+        nx = False
+        xx = False
+        while i < len(args):
+            if casematch(args[i], b'ch'):
+                ch = True
+                i += 1
+            elif casematch(args[i], b'nx'):
+                nx = True
+                i += 1
+            elif casematch(args[i], b'xx'):
+                xx = True
+                i += 1
+            else:
+                # First argument not matching flags indicates the start of
+                # score pairs.
+                break
+
+        if nx and xx:
+            raise redis.ResponseError(ZADD_NX_XX_ERROR_MSG)
+
+        elements = args[i:]
+        if not elements or len(elements) % 2 != 0:
             raise redis.ResponseError(SYNTAX_ERROR_MSG)
-        items = []
         # Parse all scores first, before updating
-        for i in range(0, len(args), 2):
-            score = Float.decode(args[i])
-            items.append((score, args[i + 1]))
+        items = [
+            (Float.decode(elements[j]), elements[j + 1])
+            for j in range(0, len(elements), 2)
+        ]
         old_len = len(zset)
-        for item in items:
-            if zset.add(item[1], item[0]):
-                key.updated()
+        changed_items = 0
+
+        for item_score, item_name in items:
+            if (
+                (not nx or item_name not in zset)
+                and (not xx or item_name in zset)
+            ):
+                if zset.add(item_name, item_score):
+                    changed_items += 1
+
+        if changed_items:
+            key.updated()
+
+        if ch:
+            return changed_items
         return len(zset) - old_len
 
     @command((Key(ZSet),))
@@ -2053,7 +2084,12 @@ class FakeSocket(object):
 
     @command((Key(ZSet), Int), (bytes, bytes))
     def zscan(self, key, cursor, *args):
-        return self._scan(key.value.items, cursor, *args)
+        new_cursor, ans = self._scan(key.value.items(), cursor, *args)
+        flat = []
+        for (key, score) in ans:
+            flat.append(key)
+            flat.append(Float.encode(score, False))
+        return [new_cursor, flat]
 
     @command((Key(ZSet), bytes))
     def zscore(self, key, member):
@@ -2107,11 +2143,19 @@ class FakeSocket(object):
             else:
                 out_members.intersection_update(s)
 
-        out = ZSet()
-        for s, w in zip(sets, weights):
+        # We first build a regular dict and turn it into a ZSet. The
+        # reason is subtle: a ZSet won't update a score from -0 to +0
+        # (or vice versa) through assignment, but a regular dict will.
+        out = {}
+        # The sort affects the order of floating-point operations.
+        # Note that redis uses qsort(1), which has no stability guarantees,
+        # so we can't be sure to match it in all cases.
+        for s, w in sorted(zip(sets, weights), key=lambda x: len(x[0])):
             for member, score in s.items():
                 score *= w
-                if math.isnan(score):
+                # Redis only does this step for ZUNIONSTORE. See
+                # https://github.com/antirez/redis/issues/3954.
+                if func == 'ZUNIONSTORE' and math.isnan(score):
                     score = 0.0
                 if member not in out_members:
                     continue
@@ -2127,10 +2171,16 @@ class FakeSocket(object):
                         score = min(old, score)
                     else:
                         assert False     # pragma: nocover
+                if math.isnan(score):
+                    score = 0.0
                 out[member] = score
 
-        dest.value = out
-        return len(out)
+        out_zset = ZSet()
+        for member, score in out.items():
+            out_zset[member] = score
+
+        dest.value = out_zset
+        return len(out_zset)
 
     @command((Key(), Int, bytes), (bytes,))
     def zunionstore(self, dest, numkeys, *args):
@@ -2185,15 +2235,15 @@ class FakeSocket(object):
 
     def _convert_redis_arg(self, lua_runtime, value):
         if isinstance(value, bytes):
-            return six.ensure_binary(value)
-        elif isinstance(value, (int, long, float)):
-            return six.ensure_binary('{:.17g}'.format(value))
+            return value
+        elif isinstance(value, (int, float)):
+            return '{:.17g}'.format(value).encode()
         else:
             # TODO: add a constant for this, and add the context
             raise redis.ResponseError('Lua redis() command arguments must be strings or integers')
 
     def _convert_redis_result(self, lua_runtime, result):
-        if isinstance(result, (bytes, int, long)):
+        if isinstance(result, (bytes, int)):
             return result
         elif isinstance(result, SimpleString):
             return lua_runtime.table_from({b"ok": result.value})
@@ -2233,8 +2283,8 @@ class FakeSocket(object):
                 item = result[index]
                 result_list.append(self._convert_lua_result(item))
             return result_list
-        elif isinstance(result, six.text_type):
-            return six.ensure_binary(result)
+        elif isinstance(result, str):
+            return result.encode()
         elif isinstance(result, float):
             return int(result)
         elif isinstance(result, bool):
@@ -2270,7 +2320,7 @@ class FakeSocket(object):
             raise redis.ResponseError(TOO_MANY_KEYS_MSG)
         if numkeys < 0:
             raise redis.ResponseError(NEGATIVE_KEYS_MSG)
-        sha1 = six.ensure_binary(hashlib.sha1(script).hexdigest())
+        sha1 = hashlib.sha1(script).hexdigest().encode()
         self._server.script_cache[sha1] = script
         lua_runtime = LuaRuntime(encoding=None, unpack_returned_tuples=True)
 
@@ -2319,7 +2369,7 @@ class FakeSocket(object):
             if len(args) != 1:
                 raise redis.ResponseError(BAD_SUBCOMMAND_MSG.format('SCRIPT'))
             script = args[0]
-            sha1 = six.ensure_binary(hashlib.sha1(script).hexdigest())
+            sha1 = hashlib.sha1(script).hexdigest().encode()
             self._server.script_cache[sha1] = script
             return sha1
         else:
@@ -2397,7 +2447,10 @@ setattr(FakeSocket, 'exec', FakeSocket.exec_)
 delattr(FakeSocket, 'exec_')
 
 
-class _DummyParser(object):
+class _DummyParser:
+    def __init__(self, socket_read_size):
+        self.socket_read_size = socket_read_size
+
     def on_disconnect(self):
         pass
 
@@ -2409,7 +2462,7 @@ class _DummyParser(object):
 try:
     from redis.selector import BaseSelector
 except ImportError:
-    class BaseSelector(object):
+    class BaseSelector:
         def __init__(self, sock):
             self.sock = sock
 
@@ -2439,15 +2492,27 @@ class FakeSelector(BaseSelector):
 class FakeConnection(redis.Connection):
     description_format = "FakeConnection<db=%(db)s>"
 
-    def __init__(self, server, db=0, password=None,
+    def __init__(self, server, db=0, username=None, password=None,
+                 socket_timeout=None, socket_connect_timeout=None,
+                 socket_keepalive=False, socket_keepalive_options=None,
+                 socket_type=0, retry_on_timeout=False,
                  encoding='utf-8', encoding_errors='strict',
-                 decode_responses=False,
-                 health_check_interval=0):
+                 decode_responses=False, parser_class=_DummyParser,
+                 socket_read_size=65536, health_check_interval=0,
+                 client_name=None):
         self.pid = os.getpid()
         self.db = db
+        self.username = username
+        self.client_name = client_name
         self.password = password
+        # Allow socket attributes to be passed in and saved even if they aren't used
+        self.socket_timeout = socket_timeout
+        self.socket_connect_timeout = socket_connect_timeout or socket_timeout
+        self.socket_keepalive = socket_keepalive
+        self.socket_keepalive_options = socket_keepalive_options or {}
+        self.socket_type = socket_type
+        self.retry_on_timeout = retry_on_timeout
         self.encoder = redis.connection.Encoder(encoding, encoding_errors, decode_responses)
-        self.retry_on_timeout = False
         self._description_args = {'db': self.db}
         self._connect_callbacks = []
         self._buffer_cutoff = 6000
@@ -2455,14 +2520,14 @@ class FakeConnection(redis.Connection):
         # self._parser isn't used for anything, but some of the
         # base class methods depend on it and it's easier not to
         # override them.
-        self._parser = _DummyParser()
+        self._parser = parser_class(socket_read_size=socket_read_size)
         self._sock = None
         # added in redis==3.3.0
         self.health_check_interval = health_check_interval
         self.next_health_check = 0
 
     def connect(self):
-        super(FakeConnection, self).connect()
+        super().connect()
         # The selector is set in redis.Connection.connect() after _connect() is called
         self._selector = FakeSelector(self._sock)
 
@@ -2504,7 +2569,7 @@ class FakeConnection(redis.Connection):
         return self._decode(response)
 
 
-class FakeRedisMixin(object):
+class FakeRedisMixin:
     def __init__(self, host='localhost', port=6379,
                  db=0, password=None, socket_timeout=None,
                  socket_connect_timeout=None,
@@ -2542,20 +2607,28 @@ class FakeRedisMixin(object):
                 'server': server
             }
             connection_pool = redis.connection.ConnectionPool(**kwargs)
-        super(FakeRedisMixin, self).__init__(
-            host, port, db, password, socket_timeout, socket_connect_timeout,
-            socket_keepalive, socket_keepalive_options, connection_pool,
-            unix_socket_path, encoding, encoding_errors, charset, errors,
-            decode_responses, retry_on_timeout,
-            ssl, ssl_keyfile, ssl_certfile, ssl_cert_reqs, ssl_ca_certs,
-            max_connections)
+        # These need to be passed by name due to
+        # https://github.com/andymccurdy/redis-py/issues/1276
+        super().__init__(
+            host=host, port=port, db=db, password=password, socket_timeout=socket_timeout,
+            socket_connect_timeout=socket_connect_timeout,
+            socket_keepalive=socket_keepalive,
+            socket_keepalive_options=socket_keepalive_options,
+            connection_pool=connection_pool,
+            unix_socket_path=unix_socket_path,
+            encoding=encoding, encoding_errors=encoding_errors,
+            charset=charset, errors=errors,
+            decode_responses=decode_responses, retry_on_timeout=retry_on_timeout,
+            ssl=ssl, ssl_keyfile=ssl_keyfile, ssl_certfile=ssl_certfile,
+            ssl_cert_reqs=ssl_cert_reqs, ssl_ca_certs=ssl_ca_certs,
+            max_connections=max_connections)
 
     @classmethod
     def from_url(cls, url, db=None, **kwargs):
         server = kwargs.pop('server', None)
         if server is None:
             server = FakeServer()
-        self = super(FakeRedisMixin, cls).from_url(url, db, **kwargs)
+        self = super().from_url(url, db, **kwargs)
         # Now override how it creates connections
         pool = self.connection_pool
         pool.connection_class = FakeConnection
